#include <math.h>
#include <xerion/useful.h>
#include <xerion/commands.h>
#include <xerion/simulator.h>

#include "unit.h"
#include "ensemble.h"
#include "ensembleCost.h"

#define M_SQR(x)		((x)*(x))

/**********************************************************************
 * If you have to have any internal state, stick it in this structure.
 * Use MCMfoo as a template on how to access the variable.
 **********************************************************************/
typedef struct EnsembleData {
    Variable	mean ;
    Variable	stdDev ;
} *EnsembleData ;

#define MCMensembleData(x,field)	\
  (((EnsembleData)((x)->costModelData.sumSquareData))->field)

#define MCMmean(x)		MCMensembleData(x,mean)
#define MCMstdDev(x)		MCMensembleData(x,stdDev)
#define MCMnet(c)		((c)->net)
#define MCMnumLinks(c)		((c)->numLinks)
#define MCMlink(c, i)		((c)->link[i])

/**********************************************************************
 * These are the procedures for getting things from the cost model:
 *
 *  MCMnet(c)		- returns the net containing the cost model
 *  MCMnumLinks(c)	- returns the number of variables in the model
 *  MCMlink(c, i) 	- returns the i'th variable ;
 *
 *  These are the procedures for manipulating variables:
 *
 *  Real variableGetValue(v) 	 - returns the value
 *  Real variableGetDeriv(v) 	 - returns the derivative
 *  char *variableGetName(v) 	 - returns its name
 *  Mask variableGetType(v)  	 - returns its type mask
 *  Net  variableGetNet(v)   	 - returns the net the variable belongs to
 *  Real variableSetValue(v,x)	 - sets the value of the variable to x
 *  Real variableSetDeriv(v,x)	 - sets the derivative of the variable to x
 *  real variableAddToDeriv(v,x) - adds x to the derivative of x
 **********************************************************************/

/**********************************************************************
 * This method should reset the cost model to its default values
 * (e.g.  it should reset means, variances, etc.)
 **********************************************************************/
static void 
reset(this, mean, stdDev)
  CostModel	this ;
  double	mean ;
  double	stdDev ;
{
  variableSetValue(MCMmean(this),   mean) ;
  variableSetValue(MCMstdDev(this), stdDev) ;
}
/**********************************************************************
 * This method should make sure that all internal variables have the
 * proper values for calculating the cost etc.
 **********************************************************************/
static void 
mysynch(this)
  CostModel     this ;
{
  if (this->missingLinks)
    MCMcompact(this) ;

  return ;
}
/**********************************************************************
 * This method should calculate the cost and return its value.
 **********************************************************************/
static Real 
ensembleCost(this)
  CostModel     this ;
{

  Real	sum ;
  Real	priorMean   = variableGetValue(MCMmean(this)) ;
  Real	priorStdDev = variableGetValue(MCMstdDev(this)) ;
  Real	priorVariance = M_SQR(priorStdDev) ;
  int	idx ;

  MCMsync(this) ;

  sum = 0.0 ;
  for (idx = 0 ; idx < MCMnumLinks(this) ; ++idx) {
    Link	link   = MCMlink(this, idx) ;
    Real	mean   = MgetWeight(link) ;
    Real	diff	 = priorMean - mean ;
    Real	variance = MgetWeightVar(link) ;
    Real	stdDev = sqrt(variance) ;

    MsetLinkCost(link,
		 this->scale*((variance - priorVariance + M_SQR(diff))
			      /(2.0*priorVariance)
			      + log(priorStdDev/stdDev))) ;
    sum += MgetLinkCost(link) ;
  }

  return sum ;
}
/**********************************************************************
 * This method should calculate the cost and add all the derivatives
 * into the list of variables.
 **********************************************************************/
static Real 
ensembleCostAndDerivs(this)
  CostModel     this ;
{
  Real	sum ;
  Real	priorMean   = variableGetValue(MCMmean(this)) ;
  Real	priorStdDev = variableGetValue(MCMstdDev(this)) ;
  Real	priorVariance = M_SQR(priorStdDev) ;
  int	idx ;

  MCMsync(this) ;

  sum = 0.0 ;
  for (idx = 0 ; idx < MCMnumLinks(this) ; ++idx) {
    Link	link	 = MCMlink(this, idx) ;
    Real	mean	 = MgetWeight(link) ;
    Real	diff	 = priorMean - mean ;
    Real	variance = MgetWeightVar(link) ;
    Real	stdDev	 = sqrt(variance) ;
    Real	ratio	 = variance/priorVariance ;

    MsetLinkCost(link,
		 this->scale*((variance - priorVariance + M_SQR(diff))
			      /(2.0*priorVariance)
			      + log(priorStdDev/stdDev))) ;
    sum += MgetLinkCost(link) ;

    MaddToWeightDeriv(link,		-this->scale*diff/priorVariance) ;
    variableAddToDeriv(MCMmean(this),	 this->scale*diff/priorVariance) ;

    MaddToWeightVarDeriv(link, this->scale*(ratio - 1.0)/(2*variance)) ;
    variableAddToDeriv(MCMstdDev(this),
		       -this->scale*(ratio - 1 + M_SQR(diff)/priorVariance)
		       /priorStdDev) ;
  }
  return sum ;
}
/**********************************************************************/

/**********************************************************************
 * DO NOT CHANGE ANYTHING AFTER THIS POINT
 **********************************************************************/

static CostModel createEnsembleCostModel ARGS((char *name, Net	net)) ;
static void	 resetEnsembleCostModel ARGS((CostModel this, double, double));

#define CM_ENSEMBLE		((int)4)

/**********************************************************************/
int
command_addEnsembleCost(tokc, tokv)
  int   tokc ;
  char *tokv[] ;
{
  CostModel	costModel ;
  Group		group ;
  String	name ;

  IUsage("[-group <group>]") ;
  if (GiveHelp(tokc)) {
    ISynopsis("set the network cost model to use ensemble cost model") ;
    IHelp
      (IHelpArgs,
       "This command sets the cost model  used by the current network to use",
       "ensemble cost model.  That is, the  network cost will be the network",
       "field \"weightCost\" times the ensemble cost (whatever that is).",
       "",
       "The \"-group\" option allows you to create a cost model that acts only",
       "on the incoming links of a  single group. If a  group does not  have",
       "it's own cost model, it uses the network cost model.",
       "",
       "SEE ALSO",
       "addMixtureCost, addSumSquareCost,",
       NULL) ;
    return 1 ;
  }

  if (currentNet == NULL)
    IErrorAbort("No currentNet") ;

  name = *tokv ;
  group = NULL ;
  for (++tokv, --tokc ; tokc > 0 ; ++tokv, --tokc) {
    if (strncmp(*tokv, "-group", strlen(*tokv)) == 0) {
      ++tokv, --tokc ;
      group = groupFromName(currentNet, *tokv) ;
      if (group == NULL)
	IErrorAbort("Unknown group: \"%s\"", *tokv) ;
    } else if (*tokv[0] == '-') {
      IErrorAbort(IPrintUsage(name, usage)) ;
    } else {
      break ;
    }
  }

  if (tokc != 0)
    IErrorAbort(IPrintUsage(name, usage)) ;

  return addEnsembleCost(currentNet, group) ;
}
/**********************************************************************/
int 
addEnsembleCost(net, group)
  Net	net ;
  Group	group ;
{
  CostModel	costModel ;

  if (group) {
    if (McostModel(group))
      MCMdestroy(McostModel(group)) ;

    McostModel(group) = createEnsembleCostModel(group->name, net) ;

  } else {
    if (McostModel(net))
      MCMdestroy(McostModel(net)) ;

    McostModel(net) = createEnsembleCostModel(net->name, net) ;
  }

  markToRebuildDisplay(ALL_DISPLAYS) ;

  return 1 ;
}
/**********************************************************************/


/*******************************************************************
 *	Name:		resetMixtureCost
 *	Description:	
 *	Parameters:
 *	  int	tokc - 
 *	  char	**tokv - 
 *	Return Value:
 *	  int	command_resetMixtureCost - 
 *******************************************************************/
int 
command_resetEnsembleCost(tokc,tokv)
  int	tokc ;
  char	*tokv[] ;
{
  CostModel	costModel ;
  Group		group ;
  String	name ;
  double	mean = 0.0 ;
  double	stdDev = 0.3 ;

  IUsage("[-group <group>] [stdDev] [mean]");
  if (GiveHelp(tokc)) {
    ISynopsis("reset the ensemble of gaussians cost model in the net");
    IHelp
      (IHelpArgs,
       "This command is used to reset the ensemble cost model in the current",
       "net (see addEnsembleCost for a description of this model).",
       "",
       "The \"-group\" option  allows you to reset the cost model of  a single",
       "group.",
       NULL) ;
    return 1 ;
  }

  if (currentNet == NULL)
    IErrorAbort("No currentNet") ;

  name = *tokv ;
  group = NULL ;
  for (++tokv, --tokc ; tokc > 0 ; ++tokv, --tokc) {
    if (strncmp(*tokv, "-group", strlen(*tokv)) == 0) {
      ++tokv, --tokc ;
      group = groupFromName(currentNet, *tokv) ;
      if (group == NULL)
	IErrorAbort("Unknown group: \"%s\"", *tokv) ;
    } else if (*tokv[0] == '-') {
      IErrorAbort(IPrintUsage(name, usage)) ;
    } else {
      break ;
    }
  }

  if (tokc > 1)
    stdDev = atof(tokv[1]) ;
  if (tokc > 2)
    mean   = atof(tokv[2]) ;
  if (tokc > 3) {
    IErrorAbort(IPrintUsage(tokv[0], usage));
    return 0 ;
  }

  if (resetEnsembleCost(currentNet, group, mean, stdDev) == 0)
    IErrorAbort("Not using ensemble of gaussians cost model") ;

  return 1 ;
}
/**********************************************************************/
int	resetEnsembleCost(net, group, mean, stdDev)
  Net	net ;
  Group	group ;
  double mean ;
  double stdDev ;
{
  CostModel	costModel ;

  if (group)
    costModel = McostModel(group) ;
  else
    costModel = McostModel(currentNet) ;

  if (costModel == NULL || MCMtype(costModel) != CM_ENSEMBLE)
    return 0 ;

  resetEnsembleCostModel(costModel, mean, stdDev) ;

  markToRebuildDisplay(ALL_DISPLAYS) ;

  return 1 ;
}
/**********************************************************************/
static void
resetEnsembleCostModel(this, mean, stdDev)
  CostModel	this ;
  double	mean ;
  double	stdDev ;
{
  if (this->type == CM_ENSEMBLE)
    this->reset(this, mean, stdDev) ;
}
/********************************************************************/


/**********************************************************************
 * This method should calculate the cost and return its value.
 **********************************************************************/
static void 
destroy(this)
  CostModel     this ;
{
  if (MCMmean(this))
    destroyVariable(MCMmean(this), MCMnet(this)) ;

  if (MCMstdDev(this))
    destroyVariable(MCMstdDev(this), MCMnet(this)) ;

  if (this->costModelData.sumSquareData)
    free((void *)this->costModelData.sumSquareData) ;
}
/**********************************************************************/
static CostModel 
createEnsembleCostModel(name, net)
  char	*name ;
  Net	net ;
{
  CostModel	this = createSumSquareCostModel(name, net) ;
  char		buffer[128] ;

  if (this == NULL)
    return NULL ;
  
  MCMtype(this) = CM_ENSEMBLE ;
  
  this->destroy			= destroy ;
  this->reset			= reset ;
  this->sync			= mysynch ;
  this->evaluateCost		= ensembleCost ;
  this->evaluateCostAndDerivs	= ensembleCostAndDerivs ;

  /* This is the messy part */
  if (this->costModelData.sumSquareData)
    free((void *)this->costModelData.sumSquareData) ;

  this->costModelData.sumSquareData = 
    (SumSquareData)malloc(sizeof(struct EnsembleData)) ;

  if (this->costModelData.sumSquareData == NULL)
    return NULL ;

  sprintf(buffer, "%s.Mean", name) ;
  MCMmean(this)	= createVariable(buffer, net, UNKNOWN) ;
  deleteLinkFromCostModel(MCMmean(this), NULL, net) ;

  sprintf(buffer, "%s.StdDeviation", name) ;
  MCMstdDev(this)= createVariable(buffer, net, LOG_TRANSFORM) ;
  deleteLinkFromCostModel(MCMstdDev(this), NULL, net) ;

  this->reset(this) ;

  return this ;
}
/**********************************************************************/
